import torch
import torch.nn as nn
import torch.nn.functional as F
from collections import OrderedDict
import numpy as np
import copy

def crt_net(hs,dropout=0.0,activation=0,bias=True,bn=0):
    net = []
    for h0,h1 in zip(hs, hs[1:]):
        if bn == 1:
            net.extend([nn.Linear(h0, h1,bias=bias),nn.BatchNorm1d(h1),nn.ReLU(),nn.Dropout(p=dropout)])
        else:
            net.extend([nn.Linear(h0, h1,bias=bias),nn.ReLU(),nn.Dropout(p=dropout)])            
    n = 2
    if bn == 1:
        n = 3
    for i in range(n):
        net.pop()
    return nn.Sequential(*net)


class NeuralNet(nn.Module):
    def __init__(self,n_att,n_out,n_hiddens=[32,32],dropout=0.0,binary=1):
        super(NeuralNet,self).__init__()
        hs = [n_att]+n_hiddens+[n_out]
        self.f = crt_net(hs,dropout=dropout)
        self.binary = binary
        
    def forward(self,x,y):
        yhat = self.f(x)
        if self.binary == 1:
            criterion = nn.BCEWithLogitsLoss()
        else:
            criterion = nn.MSELoss()
        loss = criterion(yhat,y)
        return loss

class Cate(nn.Module):
    def __init__(self,n_att,n_out=1,n_hiddens=[32,32],dropout=0.0,method=0,second_method=0):
        super(Cate,self).__init__()
        hs = [n_att]+n_hiddens+[n_out]
        self.f_pi = crt_net(hs,dropout=dropout)
        if method == 0: #sharede mu0 and mu1 S-net
            hs = [n_att]+n_hiddens
            self.f_mu = crt_net(hs,dropout=dropout)
            hs = [n_hiddens[-1]]+n_hiddens+[n_out]            
            self.f_mu0 = crt_net(hs,dropout=dropout)
            self.f_mu1 = crt_net(hs,dropout=dropout)
        elif method == 1: #separate mu0 and mu1 T-net
            hs = [n_att]+n_hiddens+[n_out]            
            self.f_mu0 = crt_net(hs,dropout=dropout)
            self.f_mu1 = crt_net(hs,dropout=dropout)
        elif method == 2: #input x and a
            hs = [n_att+1]+n_hiddens+[n_out]        
            self.f_mu_xa = crt_net(hs,dropout=dropout)

        self.method = method

    def get_mu(self,x):
        if self.method == 0:
            z_mu = self.f_mu(x)
            y0hat = self.f_mu0(z_mu)
            y1hat = self.f_mu1(z_mu)
        elif self.method == 1:
            y0hat = self.f_mu0(x)
            y1hat = self.f_mu1(x)
        elif self.method == 2:
            n_sample = x.shape[0]
            y0hat = self.f_mu_xa(torch.cat((x,torch.zeros(n_sample,1)),-1))
            y1hat = self.f_mu_xa(torch.cat((x,torch.ones(n_sample,1)),-1))
        return y0hat,y1hat
        
    def get_ytilde(self,x,y,a,second_method=0):
        pi = self.f_pi(x)
        eps = 1e-3
        pi = torch.clamp(pi,min=eps,max=1-eps)
        mu0,mu1 = self.get_mu(x)
        if second_method == 0: #RA
            ytilde = a*(y-mu0)+(1-a)*(mu1-y)
        elif second_method == 1: #PW
            ytilde = (a/pi-(1-a)/(1-pi))*y
        elif second_method == 2: #DR
            ytilde = (a/pi-(1-a)/(1-pi))*y+((1-a/pi)*mu1-(1-(1-a)/(1-pi)*mu0))
        elif second_method == 3: #no weight
            ytilde = mu1-mu0
            
        return ytilde
            

    def forward(self,x,y,a):

        a1idx = (a==1).squeeze(-1)
        a0idx = (a==0).squeeze(-1)
        y0 = y[a0idx,0]
        y1 = y[a1idx,0]

        loss = 0.0
        if y0.shape[0] > 0 and y1.shape[0]:
            ahat = self.f_pi(x)
            acriterion = nn.BCEWithLogitsLoss()
            aloss = acriterion(ahat,a)
            loss += aloss
        
        ycriterion = nn.MSELoss()
        y0hat,y1hat = self.get_mu(x)

        if y0.shape[0] > 0:            
            y0hat = y0hat[a0idx,0]
            y0loss = ycriterion(y0hat,y0)            
            loss += y0loss
        if y1.shape[0] > 0:            
            y1hat = y1hat[a1idx,0]
            y1loss = ycriterion(y1hat,y1)
            loss += y1loss
        #print("%d %d %f %f %f"%(y0.shape[0],y1.shape[0],aloss,y0loss,y1loss))
            
        return loss
        
class MetaCate(nn.Module):
    def __init__(self,n_att,n_hiddens=[32,32],dropout=0.0,method=0,second_method=0,encoder=1,GP=0,bias=0,treatment_as_feature=0):
        super(MetaCate,self).__init__()
        hs = [n_att]+n_hiddens
        if encoder > 0: #use encoder
            self.f_pi = crt_net(hs,dropout=dropout)
            if encoder == 1: #output treatment
                hs = [n_att]+n_hiddens
                self.f_mu = crt_net(hs,dropout=dropout)
            hs = [n_att]+n_hiddens            
            self.f_ytilde = crt_net(hs,dropout=dropout)

        self.log_beta = nn.Parameter(torch.randn(1))
            
        if GP > 0:
            self.log_alpha = nn.Parameter(torch.randn(1))
            self.log_gamma = nn.Parameter(torch.randn(1))
            
        self.second_method = second_method
        self.method = method
        self.encoder = encoder
        self.GP = GP #0:linear 1:GP 2:noadaptation-linear (task-shared linear)
        if GP == 2:
            self.lin_pi = nn.Linear(n_hiddens[-1],2)
            self.lin_mu0 = nn.Linear(n_hiddens[-1],1)
            self.lin_mu1 = nn.Linear(n_hiddens[-1],1)            
            self.lin_ytilde = nn.Linear(n_hiddens[-1],1)                        
        if encoder == 0:
            bias = 1
        self.bias = bias
        self.treatment_as_feature = treatment_as_feature

    def rbfkernel(self,x0,x1):
        return self.log_alpha.exp()*torch.exp(-0.5/self.log_gamma.exp()*(x0.unsqueeze(2)-x1.unsqueeze(1)).pow(2).mean(-1))
        
    def forward(self,x_spt,y_spt,a_spt,x_qry,y_qry,a_qry,tauhat_qry,tau_qry,n_spt0,n_spt1,n_qry):
        n_task,n_spt,n_att = x_spt.shape
        n_qry = x_qry.shape[1]
        n_all = n_spt+n_qry

        if self.encoder == 0:
            za_spt = x_spt
            za_qry = x_qry
        else:
            za_spt = self.f_pi(x_spt)
            za_qry = self.f_pi(x_qry)
            
        a_spt_onehot = F.one_hot(a_spt.to(torch.long))
        m = (za_spt.unsqueeze(-1)*a_spt_onehot).mean(1)
        dist = (m.unsqueeze(1)-za_qry.unsqueeze(-1)).pow(2).sum(2)
        pi_qry = torch.softmax(-dist,-1)
        dist_spt = (m.unsqueeze(1)-za_spt.unsqueeze(-1)).pow(2).sum(2)        
        pi_spt = torch.softmax(-dist_spt,-1)        

        if self.encoder == 0:
            if self.treatment_as_feature == 0:
                zy_spt = x_spt
                zy_qry = x_qry
            else:                
                zy_spt = torch.cat((x_spt,a_spt),-1)
                zy_spt0 = torch.cat((x_spt,torch.zeros_like(a_spt)),-1)
                zy_spt1 = torch.cat((x_spt,torch.ones_like(a_spt)),-1)
                zy_qry0 = torch.cat((x_qry,torch.zeros_like(a_qry)),-1)
                zy_qry1 = torch.cat((x_qry,torch.ones_like(a_qry)),-1)
                zy_qry = torch.cat((x_qry,a_qry),-1)                
        elif self.encoder == 1:
            zy_spt = self.f_mu(x_spt)
            zy_qry = self.f_mu(x_qry)

        if self.treatment_as_feature == 0:
            zy_spt0 = zy_spt[:,:n_spt0,:]
            zy_spt1 = zy_spt[:,n_spt0:,:]
        y_spt0 = y_spt[:,:n_spt0,:]
        y_spt1 = y_spt[:,n_spt0:,:]

        if self.GP == 1:
            K_spt0 = self.rbfkernel(zy_spt0,zy_spt0)+self.log_beta.exp()*torch.eye(n_spt0).unsqueeze(0).repeat(n_task,1,1).to(x_spt.device)
            K_spt1 = self.rbfkernel(zy_spt1,zy_spt1)+self.log_beta.exp()*torch.eye(n_spt1).unsqueeze(0).repeat(n_task,1,1).to(x_spt.device)
            K_qry_spt0 = self.rbfkernel(zy_qry,zy_spt0)
            K_qry_spt1 = self.rbfkernel(zy_qry,zy_spt1)
            K_spt_spt0 = self.rbfkernel(zy_spt,zy_spt0)
            K_spt_spt1 = self.rbfkernel(zy_spt,zy_spt1)
            K_spt0_inv = torch.inverse(K_spt0)
            K_spt1_inv = torch.inverse(K_spt1)
            mu0_qry = K_qry_spt0@K_spt0_inv@y_spt0
            mu1_qry = K_qry_spt1@K_spt1_inv@y_spt1
            mu0_spt = K_spt_spt0@K_spt0_inv@y_spt0
            mu1_spt = K_spt_spt1@K_spt1_inv@y_spt1
        elif self.GP == 2:
            mu0_qry = self.lin_mu0(zy_qry)
            mu1_qry = self.lin_mu1(zy_qry)
            mu0_spt = self.lin_mu0(zy_spt)
            mu1_spt = self.lin_mu1(zy_spt)
            pi_qry = torch.softmax(self.lin_pi(za_qry),-1)
            pi_spt = torch.softmax(self.lin_pi(za_spt),-1)
        else:
            if self.bias == 1:
                zy_spt0 = torch.cat((zy_spt0,torch.ones((n_task,zy_spt0.shape[1],1))),-1)
                zy_spt1 = torch.cat((zy_spt1,torch.ones((n_task,zy_spt1.shape[1],1))),-1)
                if self.treatment_as_feature == 1:
                    zy_qry0 = torch.cat((zy_qry0,torch.ones((n_task,zy_qry0.shape[1],1))),-1)
                    zy_qry1 = torch.cat((zy_qry1,torch.ones((n_task,zy_qry1.shape[1],1))),-1)
                zy_qry = torch.cat((zy_qry,torch.ones((n_task,n_qry,1))),-1)
                zy_spt = torch.cat((zy_spt,torch.ones((n_task,n_spt,1))),-1)
            n_dim = zy_spt0.shape[-1]
            if self.treatment_as_feature == 1:
                W = torch.bmm(torch.bmm(torch.inverse(torch.bmm(zy_spt.transpose(1,2),zy_spt)+self.log_beta.exp()*torch.eye(n_dim).unsqueeze(0).repeat(n_task,1,1).to(x_spt.device)),zy_spt.transpose(1,2)),y_spt)
                mu0_qry = torch.bmm(zy_qry0,W)
                mu1_qry = torch.bmm(zy_qry1,W)
                mu0_spt = torch.bmm(zy_spt0,W)
                mu1_spt = torch.bmm(zy_spt1,W)
            else:
                W0 = torch.bmm(torch.bmm(torch.inverse(torch.bmm(zy_spt0.transpose(1,2),zy_spt0)+self.log_beta.exp()*torch.eye(n_dim).unsqueeze(0).repeat(n_task,1,1).to(x_spt.device)),zy_spt0.transpose(1,2)),y_spt0)
                W1 = torch.bmm(torch.bmm(torch.inverse(torch.bmm(zy_spt1.transpose(1,2),zy_spt1)+self.log_beta.exp()*torch.eye(n_dim).unsqueeze(0).repeat(n_task,1,1).to(x_spt.device)),zy_spt1.transpose(1,2)),y_spt1)
                mu0_qry = torch.bmm(zy_qry,W0)
                mu1_qry = torch.bmm(zy_qry,W1)
                mu0_spt = torch.bmm(zy_spt,W0)
                mu1_spt = torch.bmm(zy_spt,W1)                
        mu_qry = mu0_qry*(1-a_qry)+mu1_qry*a_qry
        mu_spt = mu0_spt*(1-a_spt)+mu1_spt*a_spt

        if self.second_method == 0: #RA
            ytilde_spt = a_spt*(y_spt-mu0_spt)+(1-a_spt)*(mu1_spt-y_spt)
            ytilde_qry = a_qry*(y_qry-mu0_qry)+(1-a_qry)*(mu1_qry-y_qry)
        elif self.second_method == 1: #PW
            ytilde_spt = (a_spt/pi_spt-(1-a_spt)/(1-pi_spt))*y_spt
            ytilde_qry = (a_qry/pi_qry-(1-a_qry)/(1-pi_qry))*y_qry
        elif self.second_method == 2: #DR
            ytilde_spt = (a_spt/pi_spt-(1-a_spt)/(1-pi_spt))*y_spt+((1-a_spt/pi_spt)*mu1_spt-(1-(1-a_spt)/(1-pi_spt)*mu0_spt))
            ytilde_qry = (a_qry/pi_qry-(1-a_qry)/(1-pi_qry))*y_qry+((1-a_qry/pi_qry)*mu1_qry-(1-(1-a_qry)/(1-pi_qry)*mu0_qry))
        elif self.second_method == 3: #no weight
            ytildehat_qry = mu1_qry-mu0_qry

        if self.second_method == 0 or self.second_method == 1 or self.second_method == 2:
            if self.encoder > 0:
                zt_spt = self.f_ytilde(x_spt)
                zt_qry = self.f_ytilde(x_qry)
            else:
                zt_spt = x_spt
                zt_qry = x_qry
            if self.GP == 0:
                if self.bias == 1:
                    zt_qry = torch.cat((zt_qry,torch.ones((n_task,n_qry,1))),-1)
                    zt_spt = torch.cat((zt_spt,torch.ones((n_task,n_spt,1))),-1)                        
                n_dim = zt_spt.shape[-1]
                W = torch.bmm(torch.bmm(torch.inverse(torch.bmm(zt_spt.transpose(1,2),zt_spt)+self.log_beta.exp()*torch.eye(n_dim).unsqueeze(0).repeat(n_task,1,1).to(x_spt.device)),zt_spt.transpose(1,2)),ytilde_spt)
                ytildehat_qry = torch.bmm(zt_qry,W)
            elif self.GP == 2:
                ytildehat_qry = self.lin_ytilde(zt_qry)
            else:
                K_spt = self.rbfkernel(zt_spt,zt_spt)+self.log_beta.exp()*torch.eye(n_spt).unsqueeze(0).repeat(n_task,1,1).to(x_spt.device)
                K_qry_spt = self.rbfkernel(zt_qry,zt_spt)
                ytildehat_qry = K_qry_spt@torch.inverse(K_spt)@ytilde_spt
            
        if self.method == 0:
            loss = ((ytildehat_qry-tauhat_qry)**2).mean()
        if self.method == 1:
            piloss = (a_qry*pi_qry[:,:,[1]]+(1-a_qry)*pi_qry[:,:,[0]]).mean()
            yloss = ((mu_qry-y_qry)**2).mean()
            loss = piloss+yloss
            if self.second_method < 3:
                ytilde_qry = ytilde_qry.detach()
                ytildeloss = ((ytildehat_qry-ytilde_qry)**2).mean()
                loss += ytildeloss
            
        err = ((ytildehat_qry-tau_qry)**2).mean()        
        return loss,err

class MAMLCate(nn.Module):
    def __init__(self,n_att,n_hiddens=[32,32],dropout=0.0,second_method=0,local_epoch=5,local_lr=1e-3,method=5):
        super(MAMLCate,self).__init__()
        hs = [n_att]+n_hiddens+[2]
        self.f_pi = crt_net(hs,dropout=dropout)
        self.f_mu = crt_net(hs,dropout=dropout)
        hs = [n_att]+n_hiddens+[1]            
        self.f_ytilde = crt_net(hs,dropout=dropout)
        self.second_method = second_method
        self.local_epoch = local_epoch
        self.local_lr = local_lr
        self.method = method

    def adaptation_first(self,x,pi_weight,mu_weight):
        pi = F.relu(F.linear(x,pi_weight['0.weight'],pi_weight['0.bias']))
        pi = F.relu(F.linear(pi,pi_weight['3.weight'],pi_weight['3.bias']))
        pi = F.linear(pi,pi_weight['6.weight'],pi_weight['6.bias'])
        mu = F.relu(F.linear(x,mu_weight['0.weight'],mu_weight['0.bias']))
        mu = F.relu(F.linear(mu,mu_weight['3.weight'],mu_weight['3.bias']))
        mu = F.linear(mu,mu_weight['6.weight'],mu_weight['6.bias'])
        pi = torch.softmax(pi,-1)        
        return pi,mu

    def adaptation_second(self,x,weight):
        ytilde = F.relu(F.linear(x,weight['0.weight'],weight['0.bias']))
        ytilde = F.relu(F.linear(ytilde,weight['3.weight'],weight['3.bias']))
        ytilde = F.linear(ytilde,weight['6.weight'],weight['6.bias'])        
        return ytilde        

    def get_ytilde(self,y,a,mu,pi):
        mu0 = mu[:,:,[0]]
        mu1 = mu[:,:,[1]]
        if self.second_method == 0: #RA
            ytilde = a*(y-mu0)+(1-a)*(mu1-y)
        elif self.second_method == 1: #PW
            ytilde = (a/pi-(1-a)/(1-pi))*y
        elif self.second_method == 2: #DR
            ytilde = (a/pi-(1-a)/(1-pi))*y+((1-a/pi)*mu1-(1-(1-a)/(1-pi)*mu0))
        elif self.second_method == 3: #no weight
            ytilde = mu1-mu0
        return ytilde
        
    def forward(self,x_spt,y_spt,a_spt,x_qry,y_qry,a_qry,tauhat_qry,tau_qry,n_spt0,n_spt1,n_qry):
        n_task,n_spt,n_att = x_spt.shape
        pi_weight = OrderedDict(self.f_pi.named_parameters())
        mu_weight = OrderedDict(self.f_mu.named_parameters())
        for i in range(self.local_epoch):
            pi_spt,mu_spt = self.adaptation_first(x_spt,pi_weight,mu_weight)

            piloss = (a_spt*pi_spt[:,:,[1]]+(1-a_spt)*pi_spt[:,:,[0]]).mean()
            yloss = ((mu_spt-y_spt)**2).mean()

            pi_gradient = torch.autograd.grad(piloss,pi_weight.values(),create_graph=self.training)
            pi_weight = OrderedDict((name,param-self.local_lr*grad) for ((name,param),grad) in zip(pi_weight.items(),pi_gradient))
            mu_gradient = torch.autograd.grad(yloss,mu_weight.values(),create_graph=self.training)
            mu_weight = OrderedDict((name,param-self.local_lr*grad) for ((name,param),grad) in zip(mu_weight.items(),mu_gradient))
        if self.second_method < 3:
            ytilde_weight = OrderedDict(self.f_ytilde.named_parameters())            
            pi_spt,mu_spt = self.adaptation_first(x_spt,pi_weight,mu_weight)                
            for i in range(self.local_epoch):
                ytildehat_spt = self.adaptation_second(x_spt,ytilde_weight)
                ytilde_spt = self.get_ytilde(y_spt,a_spt,mu_spt,pi_spt).detach()                                
                ytilde_loss = ((ytildehat_spt-ytilde_spt)**2).mean()
                #print("%d %f"%(i,ytilde_loss))
                ytilde_gradient = torch.autograd.grad(ytilde_loss,ytilde_weight.values(),create_graph=self.training)
                ytilde_weight = OrderedDict((name,param-self.local_lr*grad) for ((name,param),grad) in zip(ytilde_weight.items(),ytilde_gradient))
            ytildehat_qry = self.adaptation_second(x_qry,ytilde_weight)

        pi_qry,mu_qry = self.adaptation_first(x_qry,pi_weight,mu_weight)                                
        ytilde_qry = self.get_ytilde(y_qry,a_qry,mu_qry,pi_qry).detach()
        if self.second_method == 3:
            ytildehat_qry = self.get_ytilde(y_qry,a_qry,mu_qry,pi_qry)

        if self.method == 5:
            loss = ((ytildehat_qry-tauhat_qry)**2).mean()
        elif self.method == 6:
            piloss = (a_qry*pi_qry[:,:,[1]]+(1-a_qry)*pi_qry[:,:,[0]]).mean()
            yloss = ((mu_qry-y_qry)**2).mean()
            loss = piloss+yloss
            #print("%f %f"%(piloss,yloss))
            if self.second_method < 3:
                ytildeloss = ((ytildehat_qry-ytilde_qry)**2).mean()
                #print("%f %f %f"%(piloss,yloss,ytildeloss))                
                loss += ytildeloss
            
        err = ((ytildehat_qry-tau_qry)**2).mean()        
        return loss,err


class MAMLCateIPM(nn.Module):
    def __init__(self,n_att,n_hiddens=[32,32],dropout=0.0,second_method=0,local_epoch=5,local_lr=1e-3,method=5,estimate_weight=0):
        super(MAMLCateIPM,self).__init__()
        hs = [n_att]+n_hiddens
        self.fe = crt_net(hs,dropout=dropout) #encoder
        hs = [n_hiddens[-1]]+n_hiddens+[1]        
        self.f0 = crt_net(hs,dropout=dropout) #y0|x
        self.f1 = crt_net(hs,dropout=dropout) #y1|x
        self.local_epoch = local_epoch
        self.local_lr = local_lr
        self.method = method
        self.log_alpha = nn.Parameter(torch.randn(1))
        self.log_gamma = nn.Parameter(torch.randn(1))
        self.estimate_weight = estimate_weight
        if estimate_weight == 1:
            self.log_lamb = nn.Parameter(torch.randn(1))

    def adaptationall(self,x,weight):
        z = F.relu(F.linear(x,weight['fe.0.weight'],weight['fe.0.bias']))
        z = F.linear(z,weight['fe.3.weight'],weight['fe.3.bias'])
        mu0 = F.relu(F.linear(z,weight['f0.0.weight'],weight['f0.0.bias']))
        mu0 = F.relu(F.linear(mu0,weight['f0.3.weight'],weight['f0.3.bias']))
        mu0 = F.linear(mu0,weight['f0.6.weight'],weight['f0.6.bias'])
        mu1 = F.relu(F.linear(z,weight['f1.0.weight'],weight['f1.0.bias']))
        mu1 = F.relu(F.linear(mu1,weight['f1.3.weight'],weight['f1.3.bias']))
        mu1 = F.linear(mu1,weight['f1.6.weight'],weight['f1.6.bias'])
        return z,mu0,mu1

    def adaptation(self,x,fe_weight,f0_weight,f1_weight):
        z = F.relu(F.linear(x,fe_weight['0.weight'],fe_weight['0.bias']))
        z = F.linear(z,fe_weight['3.weight'],fe_weight['3.bias'])
        mu0 = F.relu(F.linear(z,f0_weight['0.weight'],f0_weight['0.bias']))
        mu0 = F.relu(F.linear(mu0,f0_weight['3.weight'],f0_weight['3.bias']))
        mu0 = F.linear(mu0,f0_weight['6.weight'],f0_weight['6.bias'])
        mu1 = F.relu(F.linear(z,f1_weight['0.weight'],f1_weight['0.bias']))
        mu1 = F.relu(F.linear(mu1,f1_weight['3.weight'],f1_weight['3.bias']))
        mu1 = F.linear(mu1,f1_weight['6.weight'],f1_weight['6.bias'])
        return z,mu0,mu1    

    def rbfkernel(self,z0,z1):
        return self.log_alpha.exp()*torch.exp(-0.5/self.log_gamma.exp()*(z0.unsqueeze(1)-z1.unsqueeze(2))**2)
    
    def MMD(self,z0,z1):
        return self.rbfkernel(z0,z0).mean()-2*self.rbfkernel(z0,z1).mean()+self.rbfkernel(z1,z1).mean()
        
    def forward(self,x_spt,y_spt,a_spt,x_qry,y_qry,a_qry,tauhat_qry,tau_qry,n_spt0,n_spt1,n_qry):
        n_task,n_spt,n_att = x_spt.shape
        fe_weight = OrderedDict(self.fe.named_parameters())
        f0_weight = OrderedDict(self.f0.named_parameters())
        f1_weight = OrderedDict(self.f1.named_parameters())
        for i in range(self.local_epoch):
            z,mu0,mu1 = self.adaptation(x_spt,fe_weight,f0_weight,f1_weight)
            y0loss = (((1-a_spt)*(mu0-y_spt))**2).mean()
            y1loss = ((a_spt*(mu1-y_spt))**2).mean()
            z0 = z[:,:n_spt0,:]
            z1 = z[:,n_spt0:,:]
            ipmloss = self.MMD(z0,z1)
            #loss = y0loss+y1loss+100*ipmloss
            if self.estimate_weight == 1:
                ipmloss = ipmloss*self.log_lamb.exp()
                
            loss = y0loss+y1loss+ipmloss
            #print("%d %f %f %f %f"%(i,y0loss,y1loss,ipmloss,loss))
            #print(loss)
            fe_gradient = torch.autograd.grad(loss,fe_weight.values(),create_graph=self.training,retain_graph=True)
            fe_weight = OrderedDict((name,param-self.local_lr*grad) for ((name,param),grad) in zip(fe_weight.items(),fe_gradient))
            f0_gradient = torch.autograd.grad(loss,f0_weight.values(),create_graph=self.training,retain_graph=True)
            f0_weight = OrderedDict((name,param-self.local_lr*grad) for ((name,param),grad) in zip(f0_weight.items(),f0_gradient))
            f1_gradient = torch.autograd.grad(loss,f1_weight.values(),create_graph=self.training,retain_graph=True)
            f1_weight = OrderedDict((name,param-self.local_lr*grad) for ((name,param),grad) in zip(f1_weight.items(),f1_gradient))            


        z_qry,mu0_qry,mu1_qry = self.adaptation(x_qry,fe_weight,f0_weight,f1_weight)
        ytildehat_qry = mu1_qry-mu0_qry
        if self.method == 8:
            loss = ((ytildehat_qry-tauhat_qry)**2).mean()
        elif self.method == 9:
            y0loss = (((1-a_qry)*(mu0_qry-y_qry))**2).mean()
            y1loss = ((a_qry*(mu1_qry-y_qry))**2).mean()
            z0 = z_qry[:,:n_qry,:]
            z1 = z_qry[:,n_qry:,:]
            ipmloss = self.MMD(z0,z1)
            loss = y0loss+y1loss+ipmloss
            
        err = ((ytildehat_qry-tau_qry)**2).mean()

        return loss,err
    
class MultitaskCate(nn.Module):
    def __init__(self,n_task,n_att,n_hiddens=[32,32],dropout=0.0,method=0,second_method=0):
        super(MultitaskCate,self).__init__()
        hs = [n_att]+n_hiddens        
        self.f_pi = crt_net(hs,dropout=dropout)
        self.w_pi = nn.Parameter(torch.randn(n_task,n_hiddens[-1],2)) #p(a|x)
        self.f_mu = crt_net(hs,dropout=dropout)
        self.w_mu = nn.Parameter(torch.randn(n_task,n_hiddens[-1],2))
        self.f_ytilde = crt_net(hs,dropout=dropout)
        self.w_ytilde = nn.Parameter(torch.randn(n_task,n_hiddens[-1],1))
        self.second_method = second_method
        self.log_beta = nn.Parameter(torch.randn(1))
        self.beta = 1e-1
        self.bias = 1

    def forward_eval(self,x_spt,y_spt,a_spt,x_qry,y_qry,a_qry,tauhat_qry,tau_qry,n_spt0,n_spt1,n_qry):
        n_task,n_spt,n_att = x_spt.shape
        n_qry = x_qry.shape[1]
        n_all = n_spt+n_qry

        za_spt = self.f_pi(x_spt)
        za_qry = self.f_pi(x_qry)
            
        a_spt_onehot = F.one_hot(a_spt.to(torch.long))
        m = (za_spt.unsqueeze(-1)*a_spt_onehot).mean(1)
        dist = (m.unsqueeze(1)-za_qry.unsqueeze(-1)).pow(2).sum(2)
        pi_qry = torch.softmax(-dist,-1)
        dist_spt = (m.unsqueeze(1)-za_spt.unsqueeze(-1)).pow(2).sum(2)        
        pi_spt = torch.softmax(-dist_spt,-1)                

        zy_spt = self.f_mu(x_spt)
        zy_qry = self.f_mu(x_qry)

        zy_spt0 = zy_spt[:,:n_spt0,:]
        zy_spt1 = zy_spt[:,n_spt0:,:]
        y_spt0 = y_spt[:,:n_spt0,:]
        y_spt1 = y_spt[:,n_spt0:,:]

        if self.bias == 1:
            zy_spt0 = torch.cat((zy_spt0,torch.ones((n_task,n_spt0,1))),-1)
            zy_spt1 = torch.cat((zy_spt1,torch.ones((n_task,n_spt1,1))),-1)
            zy_qry = torch.cat((zy_qry,torch.ones((n_task,n_qry,1))),-1)
            zy_spt = torch.cat((zy_spt,torch.ones((n_task,n_spt,1))),-1)
            
        n_dim = zy_spt0.shape[-1]
        W0 = torch.bmm(torch.bmm(torch.inverse(torch.bmm(zy_spt0.transpose(1,2),zy_spt0)+self.beta*torch.eye(n_dim).unsqueeze(0).repeat(n_task,1,1).to(x_spt.device)),zy_spt0.transpose(1,2)),y_spt0)
        W1 = torch.bmm(torch.bmm(torch.inverse(torch.bmm(zy_spt1.transpose(1,2),zy_spt1)+self.beta*torch.eye(n_dim).unsqueeze(0).repeat(n_task,1,1).to(x_spt.device)),zy_spt1.transpose(1,2)),y_spt1)
        mu0_qry = torch.bmm(zy_qry,W0)
        mu1_qry = torch.bmm(zy_qry,W1)
        mu0_spt = torch.bmm(zy_spt,W0)
        mu1_spt = torch.bmm(zy_spt,W1)
        mu_qry = mu0_qry*(1-a_qry)+mu1_qry*a_qry
        mu_spt = mu0_spt*(1-a_spt)+mu1_spt*a_spt

        if self.second_method == 0: #RA
            ytilde_spt = a_spt*(y_spt-mu0_spt)+(1-a_spt)*(mu1_spt-y_spt)
            ytilde_qry = a_qry*(y_qry-mu0_qry)+(1-a_qry)*(mu1_qry-y_qry)
        elif self.second_method == 1: #PW
            ytilde_spt = (a_spt/pi_spt-(1-a_spt)/(1-pi_spt))*y_spt
            ytilde_qry = (a_qry/pi_qry-(1-a_qry)/(1-pi_qry))*y_qry
        elif self.second_method == 2: #DR
            ytilde_spt = (a_spt/pi_spt-(1-a_spt)/(1-pi_spt))*y_spt+((1-a_spt/pi_spt)*mu1_spt-(1-(1-a_spt)/(1-pi_spt)*mu0_spt))
            ytilde_qry = (a_qry/pi_qry-(1-a_qry)/(1-pi_qry))*y_qry+((1-a_qry/pi_qry)*mu1_qry-(1-(1-a_qry)/(1-pi_qry)*mu0_qry))
        elif self.second_method == 3: #no weight
            ytildehat_qry = mu1_qry-mu0_qry

        loss = 0.0
        if self.second_method == 0 or self.second_method == 1 or self.second_method == 2:
            zt_spt = self.f_ytilde(x_spt)
            zt_qry = self.f_ytilde(x_qry)
            n_dim = zt_spt.shape[-1]
            W = torch.bmm(torch.bmm(torch.inverse(torch.bmm(zt_spt.transpose(1,2),zt_spt)+self.beta*torch.eye(n_dim).unsqueeze(0).repeat(n_task,1,1).to(x_spt.device)),zt_spt.transpose(1,2)),ytilde_spt)
            ytildehat_qry = torch.bmm(zt_qry,W)
            ytildeloss = ((ytildehat_qry-ytilde_qry)**2).mean()
            loss += ytildeloss
            
        piloss = (a_qry*pi_qry[:,:,[1]]+(1-a_qry)*pi_qry[:,:,[0]]).mean()
        yloss = ((mu_qry-y_qry)**2).mean()
        loss += piloss+yloss
        err = ((ytildehat_qry-tau_qry)**2).mean()

        return loss,err
        
    def forward(self,x,y,a,tau,task):
        n_task,n_sample,n_att = x.shape

        za = self.f_pi(x)
        pi = za@self.w_pi[task,:,:]
        pi = torch.softmax(pi,-1)                
        
        zy = self.f_mu(x)
        mu = zy@self.w_mu[task,:,:]
        mu0 = mu[:,:,[0]]
        mu1 = mu[:,:,[1]]
        mu = mu0*(1-a)+mu1*a

        if self.second_method == 0: #RA
            ytilde = a*(y-mu0)+(1-a)*(mu1-y)
        elif self.second_method == 1: #PW
            ytilde = (a/pi-(1-a)/(1-pi))*y
        elif self.second_method == 2: #DR
            ytilde = (a/pi-(1-a)/(1-pi))*y+((1-a/pi)*mu1-(1-(1-a)/(1-pi)*mu0))
        elif self.second_method == 3: #no weight
            ytildehat = mu1-mu0

        loss = 0.0
        if self.second_method == 0 or self.second_method == 1 or self.second_method == 2:
            zt = self.f_ytilde(x)
            ytildehat = zt@self.w_ytilde[task,:,:]
            ytildeloss = ((ytildehat-ytilde)**2).mean()

        piloss = (a*pi[:,:,[1]]+(1-a)*pi[:,:,[0]]).mean()
        yloss = ((mu-y)**2).mean()
        loss += piloss+yloss
        #print("%f %f %f"%(piloss,yloss,ytildeloss))
        err = ((ytildehat-tau)**2).mean()        
        return loss,err


class SimpleCate(nn.Module):
    def __init__(self):
        super(SimpleCate,self).__init__()
    def forward(self,x_spt,y_spt,a_spt,x_qry,y_qry,a_qry,tauhat_qry,tau_qry,n_spt0,n_spt1,n_qry):
        y_spt0 = y_spt[:,:n_spt0,:]
        y_spt1 = y_spt[:,n_spt0:,:]
        ytildehat = y_spt1.mean()-y_spt0.mean()
        err = ((ytildehat-tau_qry)**2).mean()        
        return err,err
    
